; This file presents a realistic example of how easy it is
; to encounter the pitfall of the common implementations
; of delimited continuations and dynamically-scoped variables.
; The two features, as commonly implemented, interact in faulty ways.
;
; In this file, we use parameter objects (which are dynamically-scoped
; variables native to Chez Scheme) and the well-known Danvy/Filinski's
; implementation of shift/reset.
; 
; This file is (Petite) Chez Scheme code. Chez Scheme system uses parameter
; objects _extensively_,  to control many aspects of the system, from
; printing formats to garbage collection behavior. See
;	http://www.scheme.com/csug/
;	http://www.scheme.com/csug/io.html#g2021
;
; It is quite common therefore that a code that uses delimited
; continuations will also use dynvars.
;

; Our particular example uses shift/reset to traverse a tree as if it
; were a stream. Such a traversal is characteristic of the efficient
; solution to the ``same-fringe problem''. We also use the parameter
; object print-length to control the size of the printed values: ``The
; parameters print-level and print-length are useful for controlling
; the volume of output in contexts where only a small portion of the
; output is needed to identify the object being printed.''

; $Id: chez-extended-ex.scm,v 1.1 2006/03/07 07:58:14 oleg Exp $

; Danvy/Filinski's implementation of shift/reset, excerpted from Scheme48
; distribution.
(load "shift-reset.scm")		

; Convert a tree to a `stream'
;	tree->stream TREE -> STREAM
; where STREAM == () | (cons VALUE (() -> STREAM)
; As we traverse the tree, we also print the encountered values, for debugging
; purposes. 
; We must emphasize that the code tree->stream binds no dynamically
; scoped variables. Any delimited continuations captured within
; 'tree->stream' should contain no dynamic bindings.

(define (tree->stream tree)
  (reset
    (let loop ((tree tree))
      (cond
	((null? tree) '())
	((pair? tree) (loop (car tree)) (loop (cdr tree)))
	(else
	  (printf "debug output: ~s~n" tree)
	  (shift f (cons tree (lambda () (f #f))))
	  '())))))

(define tree1
  (cons (cons 1 2) (cons (vector 1 2 3 4 5) (vector 5 4 3 2 1))))

; This is a sample traversal function, which prints the tree
; depth-first.
(define (print-stream stream)
  (cond
    ((null? stream) (printf "--done--~n"))
    (else
      (printf "~s~n" (car stream))
      (print-stream ((cdr stream))))))

; This is our example. To avoid printing out the values of the tree leaves
; in full, we set print-length to abbreviate the output for vectors.

(define (test)
  (let ((stream (tree->stream tree1)))
    (parameterize ((print-length 1)) (print-stream stream))
   (display (print-length)) (newline)))

; The output is:
;
; > (test)
; debug output: 1
; 1
; debug output: 2
; 2
; debug output: #5(1 2 3 4 5)
; #5(1 ...)
; debug output: #5(5 4 3 2 1)
; #5(5 ...)
; --done--
; #f

; This shows the problem: the values output by the 'print-stream'
; procedure are indeed printed in the short form, e.g., 
; `#5(1 ...)'. OTH, the same values output by the debug print statement
; within tree->stream are printed in the full form. So, the setting of
; `print-length' to 1 had no effect on the debug print statement,
; which was certainly within the dynamic scope of the `parameterize'
; form. The observed result plainly contradicts the declarative
; semantics of dynamic variables and delimited continuations.
