; The new implementation of the parameter objects for Chez Scheme
;
; We strive to fully implement the interface of parameter objects
; in Chez Scheme. The parameter objects are _extended_ mutable 
; dynamically-scoped variables. The extension is the default value:
; if the parameter is not dynamically bound, the default value is used.
;
; $Id: new-parameters.scm,v 1.5 2006/04/10 02:25:32 oleg Exp $

; Load the implementation of the Dybvig, Sabry, Peyton-Jones' delimited
; continuation interface.
(load "cc-ref.scm")		

; Get the parameter to parameterize itself
(define-record parameterization ((immutable newv)
				 (immutable thunk)))

(define (my-make-parameter default)
  (let ((p (new-prompt)))
    (lambda args
      (cond
	((null? args)			; de-referencing
	  (if (prompt-set? p)
	    (shiftV p f (lambda (y) ((f y) y)))	; send a request
	    default))
	((null? (cdr args))
	  (cond
	    ((parameterization? (car args))
	      (dyn-bind p (parameterization-newv (car args))
		(parameterization-thunk (car args))))
	    ((prompt-set? p)		; `mutate' the dynvar, if was dyn-bound
	      (shiftV p f (lambda (y) ((f y) (car args)))))
	    (else (set! default (car args)))))
	(else
	  (error 'my-make-parameter "bad args~a~n" args))))))

(define (dyn-bind p val thunk)
  ((push-prompt p
     (let ((r (thunk)))
       (lambda (y) r)))
    val))

(define-syntax my-parameterize
  (syntax-rules ()
    ((_ ((p v)) e1 e2 ...)
      (p (make-parameterization v (lambda () e1 e2 ...))))))


;------------------------------------------------------------------------
; For simplicity and consistency with other tests, we use only one
; dynvar, so we don't bother with naming it explicitly. For
; compatibility with the SML code (see *.sml files in this directory)
; we don't make dlet a macro -- hence dlet takes a thunk as the second
; argument.


; dereference the (implicit) dynamic variable
; Just invoke (dref), which will be bound via parameterize
; (define (dref) ...)
(define dref (my-make-parameter #f))


; Evaluate THUNK in the dynamic environment extended with the binding of the
; (implicit) dynvar to V
(define (dlet v thunk)
  (my-parameterize ((dref v)) (thunk)))

; The first test
(define (test1)
  (dlet 1
    (lambda ()
      (list (dref) (dref)
      (dlet 2 
	(lambda ()
	  (list (dref) (dref))))))))

(display (test1)) (newline)
; (1 1 (2 2))


; The counter-example itself
; Its code is the same as in the other dynvar counter-examples

; A sample function to print the current dynamic environment, twice
; (just for the heck of it).
(define (tf title)
  (display title) (newline)
  (display (dref)) (newline)
  (display (dref)) (newline))

(define p0 (new-prompt))

; The first test: we capture a delimited continuation that 
; DOES include `dlet 2'
(define (test2)
  (newline) (display "test2") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
	      (push-prompt p0
		(dlet 2 
		  (lambda ()
		    ((shiftV p0 f 
		       (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test2)

; test2
; inside shift
; 1
; 1
; outside
; 1
; 1
; inside again
; 2
; 2
; outside again
; 1
; 1


; The second test: we now capture a delimited continuation that does NOT
; include `dlet 2'
(define (test3)
  (newline) (display "test3") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
		(dlet 2 
		  (lambda ()
	      (push-prompt p0
		    ((shiftV p0 f 
		       (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test3)

; test3
; inside shift
; 2
; 2
; outside
; 1
; 1
; inside again
; 1
; 1
; outside again
; 1
; 1

; Now the results are different -- and just as we expected!

;------------------------------------------------------------------------
; Run the motivating examples from Section 4 of the paper and show
; that in the present implementation, we obtain the expected results.

(printf "~nTests from the paper, Section 4~n")

(printf "Eq4, Section 4.1: ~a~n"
  (let ((p (my-make-parameter #f))
	(p0 (new-prompt)))
    (my-parameterize ((p 1))
      (push-prompt p0 (p)))))
; ==> 1

(printf "Eq5, Section 4.1: ~a~n"
  (let ((p (my-make-parameter #f))
	(p0 (new-prompt)))
    (my-parameterize ((p 1))
      (push-prompt p0 
	(my-parameterize ((p 2))
	  (shiftV p0 f (p)))))))
; ==> 1

(printf "Eq6, Section 4.2: ~a~n"
  (let ((p (my-make-parameter #f))
	(q (my-make-parameter #f))
	(p0 (new-prompt)))
    ((lambda (f)
       (my-parameterize ((p 2))
	 (my-parameterize ((q 20))
	   (f 0))))
     (my-parameterize ((p 1))
      (push-prompt p0 
	(my-parameterize ((q 10))
	  ((lambda (x) (+ (p) (q)))
	   (shiftV p0 f f))))))))
; ==> 12


;------------------------------------------------------------------------
; Now re-test  chez-extended-ex.scm
; see that file for comments

(define my-print-length (my-make-parameter #f))

; use my-parameters
(define (my-printf . args)
  (parameterize ((print-length (my-print-length)))
    (apply printf args)))
  
(define (tree->stream tree)
  (let ((p (new-prompt)))
   (push-prompt p
    (let loop ((tree tree))
      (cond
	((null? tree) '())
	((pair? tree) (loop (car tree)) (loop (cdr tree)))
	(else
	  (my-printf "debug output: ~s~n" tree)
	  (shiftV p f (cons tree (lambda () (f #f))))
	  '()))))))

(define tree1
  (cons (cons 1 2) (cons (vector 1 2 3 4 5) (vector 5 4 3 2 1))))

; This is a sample traversal function, which prints the tree
; depth-first.
(define (print-stream stream)
  (cond
    ((null? stream) (my-printf "--done--~n"))
    (else
      (my-printf "~s~n" (car stream))
      (print-stream ((cdr stream))))))

; This is our example. To avoid printing out the values of the tree leaves
; in full, we set print-length to abbreviate the output for vectors.

(my-printf "~nExtensive re-test~n")

(define (test)
  (let ((stream (tree->stream tree1)))
    (my-parameterize ((my-print-length 1)) (print-stream stream))
   (display (my-print-length)) (newline)))
(test)

; The print-out is consistent with our wishes

; Extensive re-test
; debug output: 1
; 1
; debug output: 2
; 2
; debug output: #5(1 ...)
; #5(1 ...)
; debug output: #5(5 ...)
; #5(5 ...)
; --done--
; #f

;------------------------------------------------------------------------
;   Dynamic variables and threads: we demonstrate that now we can
; select which dynvars become `private' to a thread, and which shared.
; we did not have that choice before.


(define parmp (my-make-parameter #f)) ; will be private in threads
(define parms (my-make-parameter #f)) ; this parameter will be `shared'

; A sample function to print the current dynamic environment.
(define (tf1 title)
  (printf "~a: private ~a, shared ~a~n"
    title (parmp) (parms)))

(define (test4)
  (newline) (display "test4: two threads, switching") (newline)
  (my-parameterize ((parmp 1))
    (my-parameterize ((parms 500))
      (let*
	((thread1
	   (push-prompt p0
	     (my-parameterize ((parmp (parmp))) ; this line makes parmp private
	       (tf1 "thread1: starting")
	       (my-parameterize ((parmp 10))
		   (shiftV p0 f f)		; suspend
		   (tf1 "thread1: continuing...")
		   (parmp 11) (parms 600)       ; mutate both parameters
		   (shiftV p0 f f)		; suspend
		   (tf1 "thread1: finishing, after dlet...")))))
	  (thread2
	    (my-parameterize ((parmp 2))
	      (push-prompt p0
		; the following line designates parmp private in thread2
	        (my-parameterize ((parmp (parmp))) 
		  (tf1 "thread2: starting")
		  (shiftV p0 f f)		; suspend
		  (tf1 "thread2: continuing...")
		  (my-parameterize ((parmp 22))
		    (shiftV p0 f f)		; suspend
		    (tf1 "thread2: finishing, after dlet..."))))))
	  )
	(my-parameterize ((parmp 100))
	  (my-parameterize ((parms 501))
	    (tf1 "main thread")
	    (let*
	      ((thread1 (thread1 #f))
	       (_ (tf1 "main thread"))
	       (thread2 (thread2 #f))
		)
	      (my-parameterize ((parmp 110))
		(my-parameterize ((parms 510))
		  (thread1 #f)
		  (thread2 #f)
		  )))))))))
(test4)
