{-# OPTIONS -fglasgow-exts #-}

-- In Haskell, dynamic binding is supported by the environment monad, Reader.
-- In this file, we illustrate various interactions of the environment
-- monad transformer ReaderT with the continuation monad
-- transformer ContT and the (multi-prompt) delimited continuation
-- transformer CC.
-- We observe (see the test at the end) that certain realistic idioms
-- involving delimited continuations and dynamic binding cannot be
-- expressed at all using the monad transformer mechanism because
-- no (static) combination of ReaderT and CC layers does the job.

-- We're using the delimited continuation framework by 
-- Dybvig, Peyton-Jones, and Sabry (CC monad transformers).

-- $Id: reader.hs,v 1.5 2006/05/16 22:46:16 oleg Exp $

module ReaderTCC where

import CC_FrameT -- from http://pobox.com/~oleg/ftp/packages/LogicT.tar.gz
import Control.Monad.Reader
import Control.Monad.Cont
import Control.Monad.Identity
import Control.Monad

-- Convenient (multi-prompt) control operators, in the CC monad
abortP :: Monad m => Prompt r a -> CC r m a -> CC r m b
abortP p e = letSubCont p (\_ -> e)

promptP :: Monad m => (Prompt r a -> CC r m a) -> CC r m a
promptP f = do p <- newPrompt; pushPrompt p (f p)


-- ------------------------------------------------------------------------
--                      Warm-up

-- An example of dynamic binding with the Reader monad
-- (obtained via ReaderT from the plain IO monad)
-- The Reader monad corresponds to one (implicit) dynamically-bound
-- variable. The action 'ask' dereferences this variable, the
-- action 'local modifier action1' does the dynamic binding:
-- that is, executes action1 in the changed dynamic environment.

test1 :: IO () = runReaderT (
		    do
		    task
		    local (1 +) task
		    thunk <- local (1 +) (return $ \ () -> task)
		    thunk ()
		    task
		    ) 1
    where 
    task = do { v <- ask; liftIO $ print v}

-- The output is: 1 2 1 1
-- We notice that execution of 'thunk ()' printed 1 rather than 2.
-- The 'thunk' did _not_ capture the dynamic environment
-- which was in effect at the point of its creation. This absence of
-- capture is the essence of dynamic binding.

-- The following two tests illustrate two different ways of combining
-- ReaderT and CC monad transformers. The code does no control actions:
-- we merely illustrate the combination of layers.

test2 = runCC (runReaderT action 1)
    where 
    action :: ReaderT Int (CC r IO) ()
    action = do
	     task
	     local (1 +) task
	     task
    task :: ReaderT Int (CC r IO) ()
    task = do
	   v <- ask
	   liftIO $ print v


test3 = runReaderT (runCC action) 1
    where 
    action :: CC r (ReaderT Int IO) ()
    action = do
	     lift $ task
	     promptP (\p -> lift (local (1 +) task))
	     lift $ task
    task = do { v <- ask; liftIO $ print v }


-- ------------------------------------------------------------------------
--		Combinations of ReaderT and ContT
-- Different orderings of layers give different behavior

-- In the following two examples, the action
--   callCC (\k -> local (1 +) (k 1))
-- first changes the dynamic environment and then aborts with the value 1.
-- We shall see that depending on the combination of layers, the abort
-- may fail to restore the dynamic environment changed by 'local'.

-- In this test, ReaderT is over the Cont monad
test6 = runIdentity $ runContT (runReaderT action 1) return
    where action :: ReaderT Int (ContT [Int] Identity) [Int]
          action = do
		   t1 <- ask
		   t2 <- local (1 +) ask
		   t3 <- callCC (\k -> local (1 +) ask)
		   t4 <- callCC (\k -> local (1 +) (k 1))
		   t5 <- ask
		   return [t1,t2,t3,t4,t5]
-- The output: [1,2,2,1,1]
-- What's important here is that t5 and t1 have the same value.
-- But t5 executes after t4 aborted from the local environment.
-- So, aborting restores the environment to the one existed
-- at the point of callCC


-- In this test, ContT is over the Reader monad.
-- The code for the 'action' is the same as in test6.
test7 = runIdentity $ runReaderT (runContT action return) 1
    where action :: ContT [Int] (ReaderT Int Identity) [Int]
          action = do
		   t1 <- ask
		   t2 <- local (1 +) ask
		   t3 <- callCC (\k -> local (1 +) ask)
		   t4 <- callCC (\k -> local (1 +) (k 1))
		   t5 <- ask
		   return [t1,t2,t3,t4,t5]
-- The output: [1,2,2,1,2]
-- Now, the same 'action' as in test6 gives different results:
-- t5 is 2. That is, aborting from under local (1 +)
-- has preserved that environment and failed to restore it to the
-- one existed at the point of callCC

-- An attempt to implement the following layering:
--	ReaderT ContT ReaderT
-- As we saw, the ReaderT and ContT layers do not commute, therefore,
-- we cannot merge the two ReaderT layers.
-- The two ReaderT layers in this example are equivalent to
-- two dynamic variables.

-- Alas, we cannot do any control effect from within the inner layer:
-- we get a type error if we uncomment the line "t4' <- ..." below.
-- This is the first hint that the static ordering of layers -- inherent
-- in the monad transformer approach -- lacks expressiveness.

test8 = runIdentity $ runContT (runReaderT action 1) return
    where action :: ReaderT Int (ContT [Int] Identity) [Int]
          action = do
		   t1 <- ask
		   t2 <- local (1 +) ask
		   t3 <- callCC (\k -> local (1 +) ask)
		   t4 <- callCC (\k -> local (1 +) (action1 k))
		   -- t4' <- callCC (\k -> local (1 +) (action2 k))
		   t5 <- ask
		   return [t1,t2,t3,t4,t5]
	  action1 k = do
		      runReaderT (ask >>= return . fromEnum) False
	  action2 k = do
		      runReaderT (k 3) False


-- ------------------------------------------------------------------------
--  A more elaborate example of inexpressiveness of 
--  static layering of continuation and environment monad transformers

{- 
Let us consider the following, frequently occurring pattern.  This
pattern occurs in (and is adapted from) several examples mentioned in
the paper: in equations (5) and (6), and more practically in test4 in
new-parameters.scm.  These examples are based on discussions while we
wrote the paper: http://lambda-the-ultimate.org/node/1396#comment-16007
The Zipper file-system project shows more examples of the same pattern.

    (my-parameterize ((p 500))
      (let ((thread1 
             (push-prompt q
               (let ((v1 (p)))
                 (my-parameterize ((p 600))
                   (yield q)
                   (list v1 (p)))))))
        ...))

Whereas the continuation of the first use of the parameter p includes
the prompt q, the continuation captured by (yield q) includes the
second my-parameterize.  Thus, no static layering of environment and
continuation monad transformers can implement this pattern. The
examples below show that the translation of this code into
Haskell does not type-check, no matter how the two monad transformers
are layered.
-}

-- first, we try to lift ReaderT through CC

test91 = runReaderT (runCC action) 500
    where 
    action :: CC r (ReaderT Int IO) ()
    action = do
	     lift $ task
	     -- First, we try only ReaderT effect from within promptP
	     promptP (\q -> lift (local (const 600) task))
	     -- Now, we try a CC effect from within PromptP
	     promptP (\q -> (lift task) >> abortP q (return ()))

             -- And now, we try to have both effects (as in the snippet
	     -- above)
	     -- But it doesn't type: local (const 600) body
	     -- assumes that body is ReaderT Int m
	     -- OTH, abortP is of the type CC r m ...
	     {-
	     promptP (\q -> lift (local (const 600) 
				   task >> abortP q (return ())))
	     -}
	     lift $ task
    task = do { v <- ask; liftIO $ print v }

-- Now we try to lift CC through ReaderT

test92 = runCC (runReaderT action 500)
    where 
    action :: ReaderT Int (CC r IO) ()
    action = do
	     task
	     local (1 +) task
	     -- First, we try a CC effect from within PromptP
	     v <- lift $ promptP (\q -> abortP q (return 10))
	     liftIO $ print v
	     -- Now, we try to add ReaderT effect
	     -- But it doesn't type: ask needs ReaderT, and our ReaderT
	     -- is above CC, not below CC
	     -- v <- lift $ promptP (\q -> ask >>= abortP q . return)
	     task
    task :: ReaderT Int (CC r IO) ()
    task = do
	   v <- ask
	   liftIO $ print v
